from unittest import TestCase
import os
import h5py
import numpy as np
import toml
import orcasong.tools.make_data_split as mds

__author__ = "Daniel Guderian"

test_dir = os.path.dirname(os.path.realpath(__file__))
test_data_dir = os.path.join(test_dir, "data")
# these are files that were processed with orcasong
mupage_file = os.path.join(
    test_data_dir, "processed_data_muon", "processed_graph_muon.h5"
)
neutrino_file = os.path.join(
    test_data_dir, "processed_data_neutrino", "processed_graph_neutrino.h5"
)
# config file containing 2 input groups
config_file = os.path.join(test_data_dir, "test_make_data_split_config.toml")
# the list files that will be created
list_file_dir = os.path.join(test_data_dir, "data_split_test_output", "conc_list_files")
list_output_val = os.path.join(list_file_dir, "test_list_validate_0.txt")
list_output_train = os.path.join(
    "data_split_test_output", "conc_list_files", "test_list_train_0.txt"
)
# the scripts outputs
scripts_output_dir = os.path.join(
    test_data_dir, "data_split_test_output", "job_scripts"
)
concatenate_bash_script_train = os.path.join(
    scripts_output_dir, "concatenate_h5_test_list_train_0.sh"
)
concatenate_bash_script_val = os.path.join(
    scripts_output_dir, "concatenate_h5_test_list_validate_0.sh"
)
shuffle_bash_script_train = os.path.join(
    scripts_output_dir, "shuffle_h5_test_list_train_0.sh"
)
shuffle_bash_script_val = os.path.join(
    scripts_output_dir, "shuffle_h5_test_list_validate_0.sh"
)
# and the files that will be created from these scripts
concatenate_file = os.path.join(
    "data_split_test_output", "data_split", "test_list_train_0.h5"
)


class TestMakeDataSplit(TestCase):

    """ Runs the make_data_split like in the use case. At the end, the created lists are checked."""

    @classmethod
    def setUpClass(cls):
        # the expected lists to compare to
        cls.input_categories_list = ["neutrino", "muon"]
        # include name with linebreak as they will look like this in the final files
        cls.file_path_list = [
            "processed_data_muon/processed_graph_muon.h5",
            "processed_data_neutrino/processed_graph_neutrino.h5",
            "processed_data_muon/processed_graph_muon.h5\n",
            "processed_data_neutrino/processed_graph_neutrino.h5\n",
        ]
        cls.file_path_list_val = [
            "processed_data_neutrino/processed_graph_neutrino.h5",
            "processed_data_neutrino/processed_graph_neutrino.h5\n",
        ]
        cls.n_events_list = [18, 3]
        cls.contents_concatenate_script = [
            "concatenate " + list_output_train + " --outfile " + concatenate_file
        ]
        cls.contents_shuffle_script = [
            "h5shuffle2 " + concatenate_file + " --max_ram 1000000000 \n"
        ]

        # create list_file_dir
        if not os.path.exists(list_file_dir):
            os.makedirs(list_file_dir)

    @classmethod
    def tearDownClass(cls):
        # remove the lists created
        os.remove(list_output_val)
        os.remove(list_output_train)
        os.remove(concatenate_bash_script_train)
        os.remove(concatenate_bash_script_val)
        os.remove(shuffle_bash_script_train)
        os.remove(shuffle_bash_script_val)
        os.removedirs(scripts_output_dir)
        os.removedirs(list_file_dir)
        os.removedirs(os.path.join(test_data_dir, "data_split_test_output", "logs"))
        os.removedirs(
            os.path.join(test_data_dir, "data_split_test_output", "data_split")
        )

    def test_read_keys_off_config(self):
        self.cfg = read_config(config_file)
        # get input groups and compare
        self.ip_group_keys = mds.get_all_ip_group_keys(self.cfg)
        self.assertSequenceEqual(self.ip_group_keys, self.input_categories_list)

    def test_get_filepath_and_n_events(self):
        # repeat first 2 steps
        self.cfg = read_config(config_file)
        self.ip_group_keys = mds.get_all_ip_group_keys(self.cfg)

        self.cfg, self.n_evts_total = update_cfg(self.cfg)

        for key in self.ip_group_keys:
            self.assertIn(self.cfg[key]["fpaths"][0], self.file_path_list)
            self.assertIn(self.cfg[key]["n_evts"], self.n_events_list)

    def test_make_split(self):
        # main
        # repeat first 3 steps
        self.cfg = read_config(config_file)
        self.ip_group_keys = mds.get_all_ip_group_keys(self.cfg)
        self.cfg, self.n_evts_total = update_cfg(self.cfg)

        self.cfg["n_evts_total"] = self.n_evts_total
        mds.print_input_statistics(self.cfg, self.ip_group_keys)
        for key in self.ip_group_keys:
            mds.add_fpaths_for_data_split_to_cfg(self.cfg, key)
        mds.make_dsplit_list_files(self.cfg)

        # assert the single output lists
        assert os.path.exists(list_output_val) == 1
        with open(list_output_val) as f:
            for line in f:
                self.assertIn(line, self.file_path_list_val)
        f.close

        assert os.path.exists(list_output_train) == 1
        with open(list_output_train) as f2:
            for line in f2:
                self.assertIn(line, self.file_path_list)
        f2.close

    def test_make_concatenate_and_shuffle_scripts(self):
        # main
        # repeat first 4 steps
        self.cfg = read_config(config_file)
        self.ip_group_keys = mds.get_all_ip_group_keys(self.cfg)
        self.cfg, self.n_evts_total = update_cfg(self.cfg)

        self.cfg["n_evts_total"] = self.n_evts_total
        mds.print_input_statistics(self.cfg, self.ip_group_keys)
        for key in self.ip_group_keys:
            mds.add_fpaths_for_data_split_to_cfg(self.cfg, key)
        mds.make_dsplit_list_files(self.cfg)

        # create the bash job scripts and test their content
        mds.make_concatenate_and_shuffle_scripts(self.cfg)

        assert os.path.exists(concatenate_bash_script_train) == 1
        with open(concatenate_bash_script_train) as f:
            for line in f:
                pass  # yay, awesome style! ^^
            last_line = line
            self.assertIn(last_line, self.contents_concatenate_script)
        f.close

        assert os.path.exists(shuffle_bash_script_train) == 1
        with open(shuffle_bash_script_train) as f2:
            for line in f2:
                pass
            last_line = line
            self.assertIn(last_line, self.contents_shuffle_script)
        f2.close


def update_cfg(cfg):

    """ Update the cfg with file paths and also return the total number of events"""

    # get input groups and compare
    ip_group_keys = mds.get_all_ip_group_keys(cfg)
    os.chdir(test_data_dir)
    n_evts_total = 0
    for key in ip_group_keys:
        print("Collecting information from input group " + key)
        cfg[key]["fpaths"] = mds.get_h5_filepaths(cfg[key]["dir"])
        cfg[key]["n_files"] = len(cfg[key]["fpaths"])
        (
            cfg[key]["n_evts"],
            cfg[key]["n_evts_per_file_mean"],
            cfg[key]["run_ids"],
        ) = mds.get_number_of_evts_and_run_ids(cfg[key]["fpaths"], dataset_key="y")
        n_evts_total += cfg[key]["n_evts"]

    return cfg, n_evts_total


def read_config(config_file):
    # decode config
    cfg = toml.load(config_file)
    cfg["toml_filename"] = config_file
    return cfg



from .neutrino_chain import (
    get_neutrino_mc_info_extr,
    get_muon_mc_info_extr,
    get_real_data_info_extr,
    get_random_noise_mc_info_extr,
)
from .bundles import BundleMCExtractor, BundleDataExtractor



import os
from unittest import TestCase
import tempfile
import numpy as np
import h5py
import orcasong.core
import orcasong.extractors as extractors


__author__ = "Daniel Guderian"


test_dir = os.path.dirname(os.path.realpath(__file__))
NEUTRINO_FILE = os.path.join(
    test_dir,
    "data",
    "mcv6.0.gsg_anti-muon-NC_1-500GeV.km3sim.jterbr00008350.jorcarec.jsh.aanet.898.root.h5",
)
DET_FILE_NEUTRINO = os.path.join(test_dir, "data", "KM3NeT_00000049_20200707.detx")

NOT_FULLY_RECONSTRUCTED_FILE = os.path.join(
    test_dir,
    "data",
    "mcv6.0.gsg_anti-muon-NC_1-500GeV.km3sim.jterbr00008349.jorcarec.jsh.aanet.897.root.h5",
)


class TestStdRecoExtractor(TestCase):
    """ Assert that the neutrino info is extracted correctly. File has 18 events. """

    @classmethod
    def setUpClass(cls):
        # normal case, with complete recos
        cls.proc = orcasong.core.FileGraph(
            max_n_hits=3,
            time_window=[0, 50],
            hit_infos=["pos_z", "time", "channel_id"],
            extractor=extractors.get_neutrino_mc_info_extr(NEUTRINO_FILE),
            det_file=DET_FILE_NEUTRINO,
            add_t0=True,
            keep_event_info=True,
            fixed_length=True,
        )
        cls.tmpdir = tempfile.TemporaryDirectory()
        cls.outfile = os.path.join(cls.tmpdir.name, "binned.h5")
        cls.proc.run(infile=NEUTRINO_FILE, outfile=cls.outfile)
        cls.f = h5py.File(cls.outfile, "r")
        cls.reco_names_in_original_file = ["best_jshower", "best_jmuon"]
        cls.reco_names = ["jshower", "jmuon"]
        cls.quantities_to_test = ["dir_z", "dir_x", "pos_x"]
        cls.mc_quantities_to_test = ["dir_z", "dir_x", "vertex_pos_x"]

        # case where a few recos from jmuon are missing
        cls.proc = orcasong.core.FileGraph(
            max_n_hits=3,
            time_window=[0, 50],
            hit_infos=["pos_z", "time", "channel_id"],
            extractor=extractors.get_neutrino_mc_info_extr(
                NOT_FULLY_RECONSTRUCTED_FILE
            ),
            det_file=DET_FILE_NEUTRINO,
            add_t0=True,
            keep_event_info=True,
            fixed_length=True,
        )
        cls.outfile_not_fully_rec = os.path.join(
            cls.tmpdir.name, "binned_not_fully_rec.h5"
        )
        cls.proc.run(
            infile=NOT_FULLY_RECONSTRUCTED_FILE, outfile=cls.outfile_not_fully_rec
        )
        cls.f_not_fully_rec_original = h5py.File(cls.outfile_not_fully_rec, "r")

    @classmethod
    def tearDownClass(cls):
        cls.f.close()
        cls.tmpdir.cleanup()

    def test_keys(self):
        self.assertSetEqual(
            set(self.f.keys()),
            {
                "_i_event_info",
                "_i_group_info",
                "_i_y",
                "event_info",
                "group_info",
                "x",
                "x_indices",
                "y",
            },
        )

    def test_y_contents(self):

        # the y from the processed file
        y = self.f["y"][()]

        # the info from the original file
        orig_info = h5py.File(NEUTRINO_FILE, "r")

        orig_reco_info = orig_info["reco"]
        orig_mc_info = orig_info["mc_tracks"]

        # test a few reco parameter
        for i in range(len(self.reco_names_in_original_file)):
            for j in self.quantities_to_test:
                orig_reco_values = orig_reco_info[self.reco_names_in_original_file[i]][
                    j
                ]

                assert np.allclose(
                    orig_reco_values, y[self.reco_names[i] + "_" + j], equal_nan=True
                )

        # test a few mc truth parameter
        for j in range(len(self.quantities_to_test)):
            orig_mc_values = orig_mc_info[self.quantities_to_test[j]][
                0
            ]  # only take fist, the primary

            assert np.allclose(
                orig_mc_values, y[self.mc_quantities_to_test[j]][0], equal_nan=True
            )

    def test_incomplete_reco_padding(self):

        # make sure the x and y even in this case have the same length
        x = self.f_not_fully_rec_original["x"]
        y = self.f_not_fully_rec_original["y"]

        assert len(x) == len(y)



import tempfile
import unittest
import numpy as np
import h5py
import orcasong.tools.concatenate as conc
import os

__author__ = "Stefan Reck"


class TestFileConcatenator(unittest.TestCase):
    """
    Test concatenation on pre-generated h5 files. They are in tests/data.

    """

    @classmethod
    def setUpClass(cls):
        cls.dummy_file_1 = tempfile.NamedTemporaryFile()
        _create_dummy_file(
            cls.dummy_file_1, columns=10, val_array=1, val_recarray=(1, 3)
        )
        cls.dummy_file_2 = tempfile.NamedTemporaryFile()
        _create_dummy_file(
            cls.dummy_file_2, columns=15, val_array=2, val_recarray=(4, 5)
        )
        cls.dummy_files = (
            cls.dummy_file_1.name,
            cls.dummy_file_2.name,
        )
        cls.compt_opts = {
            "complib": "gzip",
            "complevel": 1,
            "chunksize": {"numpy_array": (5, 7, 3), "rec_array": (5,)},
            "shuffle": False,
        }

    @classmethod
    def tearDownClass(cls):
        cls.dummy_file_1.close()
        cls.dummy_file_2.close()

    def test_from_list(self):
        with tempfile.NamedTemporaryFile("w+") as tf:
            tf.writelines([f + "\n" for f in self.dummy_files])
            tf.seek(0)
            fc = conc.FileConcatenator.from_list(tf.name)
            self.assertSequenceEqual(self.dummy_files, fc.input_files)

    def test_get_compopts(self):
        comptopts = conc.get_compopts(self.dummy_files[0])
        self.assertDictEqual(comptopts, self.compt_opts)

    def test_fc_get_comptopts(self):
        fc = conc.FileConcatenator(self.dummy_files)
        self.assertDictEqual(fc.comptopts, self.compt_opts)

    def test_fc_get_comptopts_updates(self):
        fc = conc.FileConcatenator(self.dummy_files, comptopts_update={"chunksize": 1})
        target_compt_opts = dict(self.compt_opts)
        target_compt_opts["chunksize"] = 1
        self.assertDictEqual(fc.comptopts, target_compt_opts)

    def test_get_cumu_rows(self):
        fc = conc.FileConcatenator(self.dummy_files)
        self.assertDictEqual(
            fc.cumu_rows, {"numpy_array": [0, 10, 25], "rec_array": [0, 10, 25]}
        )

    def test_concatenate_used_files(self):
        fc = conc.FileConcatenator(self.dummy_files)
        with tempfile.TemporaryFile() as tf:
            fc.concatenate(tf)
            with h5py.File(tf, "r") as f:
                self.assertSequenceEqual(
                    f["used_files"][()].tolist(),
                    [n.encode("ascii", "ignore") for n in self.dummy_files],
                )

    def test_concatenate_attrs(self):
        fc = conc.FileConcatenator(self.dummy_files)
        with tempfile.TemporaryFile() as tf:
            fc.concatenate(tf)
            with h5py.File(tf, "r") as f:
                target_attrs = dict(f.attrs)
                target_dset_attrs = dict(f["numpy_array"].attrs)
            with h5py.File(self.dummy_files[0], "r") as f:
                source_attrs = dict(f.attrs)
                source_dset_attrs = dict(f["numpy_array"].attrs)

            self.assertDictEqual(source_attrs, target_attrs)
            self.assertDictEqual(source_dset_attrs, target_dset_attrs)

    def test_concatenate_array(self):
        fc = conc.FileConcatenator(self.dummy_files)
        with tempfile.TemporaryFile() as tf:
            fc.concatenate(tf)
            with h5py.File(tf, "r") as f:
                target = np.ones((25, 7, 3))
                target[10:, :, :] = 2.0
                np.testing.assert_array_equal(target, f["numpy_array"][()])

    def test_concatenate_recarray_with_groupid(self):
        fc = conc.FileConcatenator(self.dummy_files)
        with tempfile.TemporaryFile() as tf:
            fc.concatenate(tf)
            with h5py.File(tf, "r") as f:
                target = np.array(
                    [(1, 3, 1)] * 25,
                    dtype=[("x", "<f8"), ("y", "<i8"), ("group_id", "<i8")],
                )
                target["x"][10:] = 4.0
                target["y"][10:] = 5.0
                target["group_id"] = np.arange(25)
                np.testing.assert_array_equal(target, f["rec_array"][()])


class BaseTestClass:
    class BaseIndexedFile(unittest.TestCase):
        @classmethod
        def setUpClass(cls) -> None:
            cls.infile = tempfile.NamedTemporaryFile()
            with h5py.File(cls.infile, "w") as f:
                cls.x = np.arange(20)
                dset_x = f.create_dataset("x", data=cls.x, chunks=True)
                dset_x.attrs.create("indexed", True)
                cls.indices = np.array(
                    [(0, 5), (5, 12), (17, 3)],
                    dtype=[("index", "<i8"), ("n_items", "<i8")],
                )
                f.create_dataset("x_indices", data=cls.indices, chunks=True)

        @classmethod
        def tearDownClass(cls) -> None:
            cls.infile.close()


class TestConcatenateIndexed(BaseTestClass.BaseIndexedFile):
    def setUp(self) -> None:
        self.outfile = "temp_out.h5"
        conc.concatenate([self.infile.name] * 2, outfile=self.outfile)

    def tearDown(self) -> None:
        if os.path.exists(self.outfile):
            os.remove(self.outfile)

    def test_check_x(self):
        with h5py.File(self.outfile) as f_out:
            np.testing.assert_array_equal(f_out["x"], np.concatenate([self.x] * 2))

    def test_check_x_indices_n_items(self):
        with h5py.File(self.outfile) as f_out:
            target_n_items = np.concatenate([self.indices] * 2)["n_items"]
            np.testing.assert_array_equal(f_out["x_indices"]["n_items"], target_n_items)

    def test_check_x_indices_index(self):
        with h5py.File(self.outfile) as f_out:
            target_n_items = np.concatenate([self.indices] * 2)["n_items"]
            target_index = np.concatenate([[0], target_n_items.cumsum()[:-1]])
            np.testing.assert_array_equal(f_out["x_indices"]["index"], target_index)

    def test_chunk_sizes_of_conc_file_are_the_same_as_in_input(self):
        with h5py.File(self.outfile) as f_out:
            self.assertTupleEqual(f_out["x"].chunks, (20,))
            self.assertTupleEqual(f_out["x_indices"].chunks, (3,))


def _create_dummy_file(filepath, columns=10, val_array=1, val_recarray=(1, 3)):
    """ Create a dummy h5 file with an array and a recarray in it. """
    with h5py.File(filepath, "w") as f:
        dset = f.create_dataset(
            "numpy_array",
            data=np.ones(shape=(columns, 7, 3)) * val_array,
            chunks=(5, 7, 3),
            compression="gzip",
            compression_opts=1,
        )
        dset.attrs.create("test_dset", "ok")

        rec_array = np.array(
            [val_recarray + (1,)] * columns,
            dtype=[("x", "<f8"), ("y", "<i8"), ("group_id", "<i8")],
        )
        rec_array["group_id"] = np.arange(columns)
        f.create_dataset(
            "rec_array",
            data=rec_array,
            chunks=(5,),
            compression="gzip",
            compression_opts=1,
        )
        f.attrs.create("test_file", "ok")



import os
import time
import datetime

import numpy as np
import psutil
import h5py
from km3pipe.sys import peak_memory_usage
import awkward as ak

from orcasong.tools.postproc import get_filepath_output, copy_used_files
from orcasong.tools.concatenate import copy_attrs


__author__ = "Stefan Reck"


def h5shuffle2(
    input_file,
    output_file=None,
    iterations=None,
    datasets=("x", "y"),
    max_ram_fraction=0.25,
    max_ram=None,
    seed=42,
):
    """
    Shuffle datasets in a h5file that have the same length.

    Parameters
    ----------
    input_file : str
        Path of the file that will be shuffle.
    output_file : str, optional
        If given, this will be the name of the output file.
        Otherwise, a name is auto generated.
    iterations : int, optional
        Shuffle the file this many times. For each additional iteration,
        a temporary file will be created and then deleted afterwards.
        Default: Auto choose best number based on available RAM.
    datasets : tuple
        Which datasets to include in output.
    max_ram : int, optional
        Available ram in bytes. Default: Use fraction of
        maximum available (see max_ram_fraction).
    max_ram_fraction : float
        in [0, 1]. Fraction of RAM to use for reading one batch of data
        when max_ram is None. Note: when using chunks, this should
        be <=~0.25, since lots of ram is needed for in-memory shuffling.
    seed : int or None
        Seed for randomness.

    Returns
    -------
    output_file : str
        Path to the output file.

    """
    if output_file is None:
        output_file = get_filepath_output(input_file, shuffle=True)
    if iterations is None:
        iterations = get_n_iterations(
            input_file,
            datasets=datasets,
            max_ram_fraction=max_ram_fraction,
            max_ram=max_ram,
        )
    # filenames of all iterations, in the right order
    filenames = (
        input_file,
        *_get_temp_filenames(output_file, number=iterations - 1),
        output_file,
    )
    if seed:
        np.random.seed(seed)
    for i in range(iterations):
        print(f"\nIteration {i+1}/{iterations}")
        _shuffle_file(
            input_file=filenames[i],
            output_file=filenames[i + 1],
            delete=i > 0,
            datasets=datasets,
            max_ram=max_ram,
            max_ram_fraction=max_ram_fraction,
        )
    return output_file


def _shuffle_file(
    input_file,
    output_file,
    datasets=("x", "y"),
    max_ram=None,
    max_ram_fraction=0.25,
    delete=False,
):
    start_time = time.time()
    if os.path.exists(output_file):
        raise FileExistsError(output_file)
    if max_ram is None:
        max_ram = get_max_ram(max_ram_fraction)
    # create file with temp name first, then rename afterwards
    temp_output_file = (
        output_file + "_temp_" + time.strftime("%d-%m-%Y-%H-%M-%S", time.gmtime())
    )
    with h5py.File(input_file, "r") as f_in:
        dsets = (*datasets, *_get_indexed_datasets(f_in, datasets))
        _check_dsets(f_in, dsets)
        dset_info = _get_largest_dset(f_in, dsets, max_ram)
        print(f"Shuffling datasets {dsets}")
        indices_per_batch = _get_indices_per_batch(
            dset_info["n_batches"],
            dset_info["n_chunks"],
            dset_info["chunksize"],
        )

        with h5py.File(temp_output_file, "x") as f_out:
            for dset_name in dsets:
                print("Creating dataset", dset_name)
                _shuffle_dset(f_out, f_in, dset_name, indices_per_batch)
                print("Done!")

    copy_used_files(input_file, temp_output_file)
    copy_attrs(input_file, temp_output_file)
    os.rename(temp_output_file, output_file)
    if delete:
        os.remove(input_file)
    print(
        f"Elapsed time: " f"{datetime.timedelta(seconds=int(time.time() - start_time))}"
    )
    return output_file


def get_max_ram(max_ram_fraction):
    max_ram = max_ram_fraction * psutil.virtual_memory().available
    print(f"Using {max_ram_fraction:.2%} of available ram = {max_ram} bytes")
    return max_ram


def get_n_iterations(
    input_file, datasets=("x", "y"), max_ram=None, max_ram_fraction=0.25
):
    """Get how often you have to shuffle with given ram to get proper randomness."""
    if max_ram is None:
        max_ram = get_max_ram(max_ram_fraction=max_ram_fraction)
    with h5py.File(input_file, "r") as f_in:
        dset_info = _get_largest_dset(f_in, datasets, max_ram)
    n_iterations = np.amax(
        (
            1,
            int(
                np.ceil(
                    np.log(dset_info["n_chunks"])
                    / np.log(dset_info["chunks_per_batch"])
                )
            ),
        )
    )
    print(f"Largest dataset: {dset_info['name']}")
    print(f"Total chunks: {dset_info['n_chunks']}")
    print(f"Max. chunks per batch: {dset_info['chunks_per_batch']}")
    print(f"--> min iterations for full shuffle: {n_iterations}")
    return n_iterations


def _get_indices_per_batch(n_batches, n_chunks, chunksize):
    """
    Return a list with the shuffled indices for each batch.

    Returns
    -------
    indices_per_batch : List
        Length n_batches, each element is a np.array[int].
        Element i of the list are the indices of each sample in batch number i.

    """
    chunk_indices = np.arange(n_chunks)
    np.random.shuffle(chunk_indices)
    chunk_batches = np.array_split(chunk_indices, n_batches)

    indices_per_batch = []
    for bat in chunk_batches:
        idx = (bat[:, None] * chunksize + np.arange(chunksize)[None, :]).flatten()
        np.random.shuffle(idx)
        indices_per_batch.append(idx)

    return indices_per_batch


def _get_largest_dset(f, datasets, max_ram):
    """
    Get infos about the dset that needs the most batches.
    This is the dset that determines how many samples are shuffled at a time.
    """
    dset_infos = _get_dset_infos(f, datasets, max_ram)
    return dset_infos[np.argmax([v["n_batches"] for v in dset_infos])]


def _check_dsets(f, datasets):
    # check if all datasets have the same number of lines
    n_lines_list = []
    for dset_name in datasets:
        if dset_is_indexed(f, dset_name):
            dset_name = f"{dset_name}_indices"
        n_lines_list.append(len(f[dset_name]))

    if not all([n == n_lines_list[0] for n in n_lines_list]):
        raise ValueError(f"Datasets have different lengths! " f"{n_lines_list}")


def _get_indexed_datasets(f, datasets):
    indexed_datasets = []
    for dset_name in datasets:
        if dset_is_indexed(f, dset_name):
            indexed_datasets.append(f"{dset_name}_indices")
    return indexed_datasets


def _get_dset_infos(f, datasets, max_ram):
    """Retrieve infos for each dataset."""
    dset_infos = []
    for i, name in enumerate(datasets):
        if name.endswith("_indices"):
            continue
        if dset_is_indexed(f, name):
            # for indexed dataset: take average bytes in x per line in x_indices
            dset_data = f[name]
            name = f"{name}_indices"
            dset = f[name]
            bytes_per_line = np.asarray(dset[0]).nbytes * len(dset_data) / len(dset)
        else:
            dset = f[name]
            bytes_per_line = np.asarray(dset[0]).nbytes

        n_lines = len(dset)
        chunksize = dset.chunks[0]
        n_chunks = int(np.ceil(n_lines / chunksize))
        bytes_per_chunk = bytes_per_line * chunksize
        chunks_per_batch = int(np.floor(max_ram / bytes_per_chunk))

        dset_infos.append(
            {
                "name": name,
                "n_chunks": n_chunks,
                "chunks_per_batch": chunks_per_batch,
                "n_batches": int(np.ceil(n_chunks / chunks_per_batch)),
                "chunksize": chunksize,
            }
        )

    return dset_infos


def dset_is_indexed(f, dset_name):
    if f[dset_name].attrs.get("indexed"):
        if f"{dset_name}_indices" not in f:
            raise KeyError(
                f"{dset_name} is indexed, but {dset_name}_indices is missing!"
            )
        return True
    else:
        return False


def _shuffle_dset(f_out, f_in, dset_name, indices_per_batch):
    """
    Create a batchwise-shuffled dataset in the output file using given indices.

    """
    dset_in = f_in[dset_name]
    start_idx = 0
    for batch_number, indices in enumerate(indices_per_batch):
        print(f"Processing batch {batch_number+1}/{len(indices_per_batch)}")
        # remove indices outside of dset
        if dset_is_indexed(f_in, dset_name):
            max_index = len(f_in[f"{dset_name}_indices"])
        else:
            max_index = len(dset_in)
        indices = indices[indices < max_index]

        # reading has to be done with linearly increasing index
        #  fancy indexing is super slow
        #  so sort -> turn to slices -> read -> conc -> undo sorting
        sort_ix = np.argsort(indices)
        unsort_ix = np.argsort(sort_ix)
        fancy_indices = indices[sort_ix]
        slices = _slicify(fancy_indices)

        if dset_is_indexed(f_in, dset_name):
            # special treatment for indexed: slice based on indices dataset
            slices_indices = [f_in[f"{dset_name}_indices"][slc] for slc in slices]
            data = np.concatenate(
                [dset_in[slice(*_resolve_indexed(slc))] for slc in slices_indices]
            )
            # convert to 3d awkward array, then shuffle, then back to numpy
            data_indices = np.concatenate(slices_indices)
            data_ak = ak.unflatten(data, data_indices["n_items"])
            data = ak.flatten(data_ak[unsort_ix], axis=1).to_numpy()

        else:
            data = np.concatenate([dset_in[slc] for slc in slices])
            data = data[unsort_ix]

        if dset_name.endswith("_indices"):
            # recacalculate index
            data["index"] = start_idx + np.concatenate(
                [[0], np.cumsum(data["n_items"][:-1])]
            )

        if batch_number == 0:
            out_dset = f_out.create_dataset(
                dset_name,
                data=data,
                maxshape=dset_in.shape,
                chunks=dset_in.chunks,
                compression=dset_in.compression,
                compression_opts=dset_in.compression_opts,
                shuffle=dset_in.shuffle,
            )
            out_dset.resize(len(dset_in), axis=0)
            start_idx = len(data)
        else:
            end_idx = start_idx + len(data)
            f_out[dset_name][start_idx:end_idx] = data
            start_idx = end_idx

        print("Memory peak: {0:.3f} MB".format(peak_memory_usage()))

    if start_idx != len(dset_in):
        print(f"Warning: last index was {start_idx} not {len(dset_in)}")


def _slicify(fancy_indices):
    """[0,1,2, 6,7,8] --> [0:3, 6:9]"""
    steps = np.diff(fancy_indices) != 1
    slice_starts = np.concatenate([fancy_indices[:1], fancy_indices[1:][steps]])
    slice_ends = np.concatenate([fancy_indices[:-1][steps], fancy_indices[-1:]]) + 1
    return [slice(slice_starts[i], slice_ends[i]) for i in range(len(slice_starts))]


def _resolve_indexed(ind):
    # based on slice of x_indices, get where to slice in x
    return ind["index"][0], ind["index"][-1] + ind["n_items"][-1]


def _get_temp_filenames(output_file, number):
    path, file = os.path.split(output_file)
    return [os.path.join(path, f"temp_iteration_{i}_{file}") for i in range(number)]


def run_parser():
    # TODO deprecated
    raise NotImplementedError("h5shuffle2 has been renamed to orcasong h5shuffle2")



import os
from unittest import TestCase
import tempfile
import numpy as np
import h5py
import orcasong.core
import orcasong.extractors as extractors
from orcasong.plotting.plot_binstats import read_hists_from_h5file


__author__ = "Stefan Reck"


test_dir = os.path.dirname(os.path.realpath(__file__))
MUPAGE_FILE = os.path.join(test_dir, "data", "mupage.root.h5")
DET_FILE = os.path.join(test_dir, "data", "KM3NeT_-00000001_20171212.detx")


class TestFileBinner(TestCase):
    """ Assert that the filebinner still produces the same output. """

    @classmethod
    def setUpClass(cls):
        cls.proc = orcasong.core.FileBinner(
            bin_edges_list=[
                ["pos_z", np.linspace(0, 200, 3)],
                ["time", np.linspace(0, 600, 3)],
                ["channel_id", np.linspace(-0.5, 30.5, 3)],
            ],
            extractor=extractors.get_real_data_info_extr(MUPAGE_FILE),
            det_file=DET_FILE,
            add_t0=True,
            keep_event_info=True,
        )
        cls.tmpdir = tempfile.TemporaryDirectory()
        cls.outfile = os.path.join(cls.tmpdir.name, "binned.h5")
        cls.proc.run(infile=MUPAGE_FILE, outfile=cls.outfile)
        cls.f = h5py.File(cls.outfile, "r")

    @classmethod
    def tearDownClass(cls):
        cls.f.close()
        cls.tmpdir.cleanup()

    def test_keys(self):
        self.assertSetEqual(
            set(self.f.keys()),
            {
                "_i_event_info",
                "_i_group_info",
                "_i_y",
                "bin_stats",
                "event_info",
                "group_info",
                "x",
                "x_indices",
                "y",
            },
        )

    def test_x(self):
        target = np.array(
            [
                [[[4, 1], [6, 3]], [[12, 5], [6, 7]]],
                [[[4, 2], [1, 3]], [[5, 7], [8, 5]]],
                [[[3, 3], [2, 4]], [[5, 6], [6, 8]]],
            ],
            dtype=np.uint8,
        )
        np.testing.assert_equal(target, self.f["x"])

    def test_y(self):
        y = self.f["y"][()]
        target = {
            "event_id": np.array([0.0, 1.0, 2.0]),
            "run_id": np.array([1.0, 1.0, 1.0]),
            "trigger_mask": np.array([18.0, 18.0, 16.0]),
            "group_id": np.array([0, 1, 2]),
        }
        for k, v in target.items():
            np.testing.assert_equal(y[k], v)

    def test_bin_stats(self):
        bin_stats = read_hists_from_h5file(self.f)
        target_hists = {
            "channel_id": np.array(
                [20.0, 8.0, 10.0, 8.0, 16.0, 13.0, 11.0, 9.0, 10.0, 11.0]
            ),
            "pos_z": np.array([0.0, 0.0, 0.0, 36.0, 0.0, 19.0, 30.0, 0.0, 31.0, 0.0]),
            "time": np.array([57.0, 59.0]),
        }
        for dim, infos in bin_stats.items():
            np.testing.assert_equal(infos["hist"], target_hists[dim])


class TestFileGraph(TestCase):
    """ Assert that the FileGraph still produces the same output. """

    @classmethod
    def setUpClass(cls):
        # produce test file, once for fixed_length (old format), and once
        # for the new format
        cls.proc_fixed_length, cls.proc = [
            orcasong.core.FileGraph(
                max_n_hits=3,
                time_window=[0, 50],
                hit_infos=["pos_z", "time", "channel_id"],
                extractor=extractors.get_real_data_info_extr(MUPAGE_FILE),
                det_file=DET_FILE,
                add_t0=False,
                keep_event_info=True,
                correct_timeslew=False,
                fixed_length=fixed_length,
            )
            for fixed_length in (True, False)
        ]
        cls.tmpdir = tempfile.TemporaryDirectory()

        cls.outfile_fixed_length = os.path.join(
            cls.tmpdir.name, "binned_fixed_length.h5"
        )
        cls.proc_fixed_length.run(infile=MUPAGE_FILE, outfile=cls.outfile_fixed_length)
        cls.f_fixed_length = h5py.File(cls.outfile_fixed_length, "r")

        cls.outfile = os.path.join(cls.tmpdir.name, "binned.h5")
        cls.proc.run(infile=MUPAGE_FILE, outfile=cls.outfile)
        cls.f = h5py.File(cls.outfile, "r")

    @classmethod
    def tearDownClass(cls):
        cls.f.close()
        cls.f_fixed_length.close()
        cls.tmpdir.cleanup()

    def test_keys_fixed_length(self):
        self.assertSetEqual(
            set(self.f_fixed_length.keys()),
            {
                "_i_event_info",
                "_i_group_info",
                "_i_y",
                "event_info",
                "group_info",
                "x",
                "x_indices",
                "y",
            },
        )

    def test_keys(self):
        self.assertSetEqual(
            set(self.f_fixed_length.keys()),
            {
                "_i_event_info",
                "_i_group_info",
                "_i_y",
                "event_info",
                "group_info",
                "x",
                "x_indices",
                "y",
            },
        )

    def test_x_attrs_fixed_length(self):
        to_check = {
            "hit_info_0": "pos_z",
            "hit_info_1": "time",
            "hit_info_2": "channel_id",
            "hit_info_3": "is_valid",
            "indexed": False,
        }
        attrs = dict(self.f_fixed_length["x"].attrs)
        for k, v in to_check.items():
            self.assertTrue(attrs[k] == v)

    def test_x_attrs(self):
        to_check = {
            "hit_info_0": "pos_z",
            "hit_info_1": "time",
            "hit_info_2": "channel_id",
            "indexed": True,
        }
        attrs = dict(self.f["x"].attrs)
        for k, v in to_check.items():
            self.assertTrue(attrs[k] == v)

    def test_x_fixed_length(self):
        target = np.array(
            [
                [
                    [676.941, 13.0, 30.0, 1.0],
                    [461.111, 32.0, 9.0, 1.0],
                    [424.941, 1.0, 30.0, 1.0],
                ],
                [
                    [172.83, 32.0, 25.0, 1.0],
                    [316.83, 2.0, 14.0, 1.0],
                    [461.059, 1.0, 3.0, 1.0],
                ],
                [
                    [496.83, 34.0, 25.0, 1.0],
                    [605.111, 9.0, 4.0, 1.0],
                    [424.889, 46.0, 29.0, 1.0],
                ],
            ],
            dtype=np.float32,
        )
        np.testing.assert_equal(target, self.f_fixed_length["x"])

    def test_x(self):
        target = np.array(
            [
                [676.941, 13.0, 30.0],
                [461.111, 32.0, 9.0],
                [424.941, 1.0, 30.0],
                [172.83, 32.0, 25.0],
                [316.83, 2.0, 14.0],
                [461.059, 1.0, 3.0],
                [496.83, 34.0, 25.0],
                [605.111, 9.0, 4.0],
                [424.889, 46.0, 29.0],
            ],
            dtype=np.float32,
        )
        np.testing.assert_equal(target, self.f["x"])

    def test_y_fixed_length(self):
        y = self.f_fixed_length["y"][()]
        target = {
            "event_id": np.array([0.0, 1.0, 2.0]),
            "run_id": np.array([1.0, 1.0, 1.0]),
            "trigger_mask": np.array([18.0, 18.0, 16.0]),
            "group_id": np.array([0, 1, 2]),
        }
        for k, v in target.items():
            np.testing.assert_equal(y[k], v)

    def test_y(self):
        y = self.f["y"][()]
        target = {
            "event_id": np.array([0.0, 1.0, 2.0]),
            "run_id": np.array([1.0, 1.0, 1.0]),
            "trigger_mask": np.array([18.0, 18.0, 16.0]),
            "group_id": np.array([0, 1, 2]),
        }
        for k, v in target.items():
            np.testing.assert_equal(y[k], v)



import warnings
import numpy as np


class BundleDataExtractor:
    """Get info present in real data."""

    def __init__(self, infile, only_downgoing_tracks=False):
        self.only_downgoing_tracks = only_downgoing_tracks

    def __call__(self, blob):
        # just take everything from event info
        if not len(blob["EventInfo"]) == 1:
            warnings.warn(f"Event info has length {len(blob['EventInfo'])}, not 1")
        track = dict(zip(blob["EventInfo"].dtype.names, blob["EventInfo"][0]))
        track.update(
            **get_best_track(blob, only_downgoing_tracks=self.only_downgoing_tracks)
        )

        track["n_hits"] = len(blob["Hits"])
        track["n_triggered_hits"] = blob["Hits"]["triggered"].sum()
        is_triggered = blob["Hits"]["triggered"].astype(bool)
        track["n_triggered_doms"] = len(np.unique(blob["Hits"]["dom_id"][is_triggered]))
        track["t_last_triggered"] = blob["Hits"]["time"][is_triggered].max()

        unique_hits = get_only_first_hit_per_pmt(blob["Hits"])
        track["n_pmts"] = len(unique_hits)
        track["n_triggered_pmts"] = unique_hits["triggered"].sum()

        if "n_hits_intime" in blob["EventInfo"]:
            n_hits_intime = blob["EventInfo"]["n_hits_intime"]
        else:
            n_hits_intime = np.nan
        track["n_hits_intime"] = n_hits_intime
        return track


def get_only_first_hit_per_pmt(hits):
    """Keep only the first hit of each pmt."""
    idents = np.stack((hits["dom_id"], hits["channel_id"]), axis=-1)
    sorted_time_indices = np.argsort(hits["time"])
    # indices of first hit per pmt in time sorted array:
    indices = np.unique(idents[sorted_time_indices], axis=0, return_index=True)[1]
    # indices of first hit per pmt in original array:
    first_hit_indices = np.sort(sorted_time_indices[indices])
    return hits[first_hit_indices]


def get_best_track(blob, missing_value=np.nan, only_downgoing_tracks=False):
    """
    I mean first track, i.e. the one with longest chain and highest lkl/nhits.
    Can also take the best track only of those that are downgoing.
    """
    # hardcode names here since the first blob might not have Tracks
    names = (
        "E",
        "JCOPY_Z_M",
        "JENERGY_CHI2",
        "JENERGY_ENERGY",
        "JENERGY_MUON_RANGE_METRES",
        "JENERGY_NDF",
        "JENERGY_NOISE_LIKELIHOOD",
        "JENERGY_NUMBER_OF_HITS",
        "JGANDALF_BETA0_RAD",
        "JGANDALF_BETA1_RAD",
        "JGANDALF_CHI2",
        "JGANDALF_LAMBDA",
        "JGANDALF_NUMBER_OF_HITS",
        "JGANDALF_NUMBER_OF_ITERATIONS",
        "JSHOWERFIT_ENERGY",
        "JSTART_LENGTH_METRES",
        "JSTART_NPE_MIP",
        "JSTART_NPE_MIP_TOTAL",
        "JVETO_NPE",
        "JVETO_NUMBER_OF_HITS",
        "dir_x",
        "dir_y",
        "dir_z",
        "id",
        "length",
        "likelihood",
        "pos_x",
        "pos_y",
        "pos_z",
        "rec_type",
        "t",
        "group_id",
    )
    if "Tracks" in blob:
        tracks = blob["Tracks"]
    elif "BestJmuon" in blob:
        if only_downgoing_tracks:
            raise ValueError("only_downgoing_tracks option requires all tracks!")
        tracks = blob["BestJmuon"]
    else:
        tracks = None

    index = None
    if tracks is not None:
        if only_downgoing_tracks:
            downs = np.where(tracks.dir_z < 0)[0]
            if len(downs) != 0:
                index = downs[0]
        else:
            index = 0

    if index is not None:
        track = tracks[index]
        return {f"jg_{name}_reco": track[name] for name in names}
    else:
        return {f"jg_{name}_reco": missing_value for name in names}


class BundleMCExtractor:
    """
    For atmospheric muon studies on mupage or corsika simulations.

    Parameters
    ----------
    inactive_du : int, optional
        Don't count mchits in this du. E.g. for ORCA4, DU 1 is inactive.
    min_n_mchits_list : tuple
        How many mchits does a muon have to produce to be counted?
        Create a seperate set of entries for each number in the tuple.
    plane_point : tuple
        For bundle diameter: XYZ coordinates of where the center of the
        plane is in which the muon positions get calculated. Should be set
        to the center of the detector!
    with_mc_index : bool
        Add a column called mc_index containing the mc run number,
        which is attempted to be read from the filename. This is for
        when the same run id/event id combination appears in mc files,
        which can happend e.g. in run by run simulations when there are
        multiplie mc runs per data run.
        Requires the filename to have a very specific format, which is
        likely not future-proof.
        TODO this would ideally not be read from the filename,
         but there is currently not other way of accessing it (07/2021).
    is_corsika : bool
        Use this when using Corsika!!!
    only_downgoing_tracks : bool
        For the best track (JG reco), consider only the ones that are downgoing.
    missing_value : float
        If a value is missing, use this value instead.

    """

    def __init__(
        self,
        infile,
        inactive_du=None,
        min_n_mchits_list=(0, 1, 10),
        plane_point=(17, 17, 111),
        with_mc_index=True,
        is_corsika=False,
        only_downgoing_tracks=False,
        missing_value=np.nan,
    ):
        self.inactive_du = inactive_du
        self.min_n_mchits_list = min_n_mchits_list
        self.plane_point = plane_point
        self.with_mc_index = with_mc_index
        self.missing_value = missing_value
        self.is_corsika = is_corsika
        self.only_downgoing_tracks = only_downgoing_tracks

        self.data_extractor = BundleDataExtractor(
            infile, only_downgoing_tracks=only_downgoing_tracks
        )

        if self.with_mc_index:
            self.mc_index = get_mc_index(infile)
            print(f"Using mc_index {self.mc_index}")
        else:
            self.mc_index = None

    def __call__(self, blob):
        mc_info = self.data_extractor(blob)

        if self.is_corsika:
            # Corsika has a primary particle. Store infos about it
            prim_track = blob["McTracks"][0]

            # primary should be track 0 with id 0
            if prim_track["id"] != 0:
                raise ValueError("Error finding primary: mc_tracks[0]['id'] != 0")

            # direction of the primary
            mc_info["dir_x"] = prim_track.dir_x
            mc_info["dir_y"] = prim_track.dir_y
            mc_info["dir_z"] = prim_track.dir_z
            # use primary direction as plane normal
            plane_normal = np.array(prim_track[["dir_x", "dir_y", "dir_z"]].tolist())

            for fld in ("pos_x", "pos_y", "pos_z", "pdgid", "energy", "time"):
                mc_info[f"primary_{fld}"] = prim_track[fld]

            # remove primary for the following, since it's not a muon
            blob["McTracks"] = blob["McTracks"][1:]
        else:
            # In mupage, all muons in a bundle are parallel. So just take dir of first muon
            mc_info["dir_x"] = blob["McTracks"].dir_x[0]
            mc_info["dir_y"] = blob["McTracks"].dir_y[0]
            mc_info["dir_z"] = blob["McTracks"].dir_z[0]
            plane_normal = None

        # n_mc_hits of each muon in active dus
        mchits_per_muon = get_mchits_per_muon(blob, inactive_du=self.inactive_du)

        for min_n_mchits in self.min_n_mchits_list:
            if min_n_mchits == 0:
                mc_tracks_sel = blob["McTracks"]
                suffix = "sim"
            else:
                mc_tracks_sel = blob["McTracks"][mchits_per_muon >= min_n_mchits]
                suffix = f"{min_n_mchits}_mchits"

            # total number of mchits of all muons
            mc_info[f"n_mc_hits_{suffix}"] = np.sum(
                mchits_per_muon[mchits_per_muon >= min_n_mchits]
            )
            # number of muons with at least the given number of mchits
            mc_info[f"n_muons_{suffix}"] = len(mc_tracks_sel)
            # summed up energy of all muons
            mc_info[f"energy_{suffix}"] = np.sum(mc_tracks_sel.energy)
            # bundle diameter; only makes sense for 2+ muons
            if len(mc_tracks_sel) >= 2:
                positions_plane = get_plane_positions(
                    positions=mc_tracks_sel[["pos_x", "pos_y", "pos_z"]]
                    .to_dataframe()
                    .to_numpy(),
                    directions=mc_tracks_sel[["dir_x", "dir_y", "dir_z"]]
                    .to_dataframe()
                    .to_numpy(),
                    plane_point=self.plane_point,
                    plane_normal=plane_normal,
                )
                pairwise_distances = get_pairwise_distances(positions_plane)
                mc_info[f"max_pair_dist_{suffix}"] = pairwise_distances.max()
                mc_info[f"mean_pair_dist_{suffix}"] = pairwise_distances.mean()
            else:
                mc_info[f"max_pair_dist_{suffix}"] = self.missing_value
                mc_info[f"mean_pair_dist_{suffix}"] = self.missing_value

        if self.with_mc_index:
            mc_info["mc_index"] = self.mc_index

        return mc_info


def get_plane_positions(positions, directions, plane_point, plane_normal=None):
    """
    Get the position of each muon in a 2d plane.
    Length will be preserved, i.e. 1m in 3d space is also 1m in plane space.

    Parameters
    ----------
    positions : np.array
        The position of each muon in 3d cartesian space, shape (n_muons, 3).
    directions : np.array
        The direction of each muon as a cartesian unit vector, shape (n_muons, 3).
    plane_point : np.array
        A 3d cartesian point on the plane. This will be (0, 0) in the plane
        coordinate system. Shape (3, ).
    plane_normal : np.array, optional
        A 3d cartesian vector perpendicular to the plane, shape (3, ).
        Default: Use directions if all muons are parallel, otherwise raise.

    Returns
    -------
    positions_plane : np.array
        The 2d position of each muon in the plane, shape (n_muons, 2).

    """
    if plane_normal is None:
        if not np.all(directions == directions[0]):
            raise ValueError(
                "Muon tracks are not all parallel: plane_normal has to be specified!"
            )
        plane_normal = directions[0]

    # get the 3d points where each muon collides with the plane
    points = []
    for i in range(len(directions)):
        ndotu = np.dot(plane_normal, directions[i])
        if abs(ndotu) < 1e-6:
            raise ValueError("no intersection or line is within plane")

        w = positions[i] - plane_point
        si = -np.dot(plane_normal, w) / ndotu
        psi = w + si * directions[i] + plane_point
        points.append(psi)
    points = np.array(points)

    # Get the unit vectors of the plane. u is 0 in x, v is 0 in y.
    u = np.array([1, 0, -plane_normal[0] / plane_normal[2]])
    v = np.array([0, 1, -plane_normal[1] / plane_normal[2]])
    # norm:
    u = u / np.linalg.norm(u)
    v = v / np.linalg.norm(v)

    # xy coordinates in plane
    x_dash = (points[:, 0] - plane_point[0]) / u[0]
    y_dash = (points[:, 1] - plane_point[1]) / v[1]
    position_plane = np.array([x_dash, y_dash]).T

    return position_plane


def get_pairwise_distances(positions_plane, as_matrix=False):
    """
    Get the perpendicular distance between each muon pair.

    Parameters
    ----------
    positions_plane : np.array
        The 2d position of each muon in a plane, shape (n_muons, 2).
    as_matrix : bool
        Return the whole 2D distance matrix.

    Returns
    -------
    np.array
        The distances between each pair of muons.
        1D if as_matrix is False (default), else 2D.

    """
    pos_x, pos_y = positions_plane[:, 0], positions_plane[:, 1]

    dists_x = np.expand_dims(pos_x, -2) - np.expand_dims(pos_x, -1)
    dists_y = np.expand_dims(pos_y, -2) - np.expand_dims(pos_y, -1)
    l2_dists = np.sqrt(dists_x ** 2 + dists_y ** 2)
    if as_matrix:
        return l2_dists
    else:
        return l2_dists[np.triu_indices_from(l2_dists, k=1)]


def get_mchits_per_muon(blob, inactive_du=None):
    """
    For each muon in McTracks, get the number of McHits.

    Parameters
    ----------
    blob
        The blob.
    inactive_du : int, optional
        McHits in this DU will not be counted.

    Returns
    -------
    np.array
        n_mchits, len = number of muons --> blob["McTracks"]["id"]

    """
    ids = blob["McTracks"]["id"]
    # Origin of each mchit (as int) in the active line
    origin = blob["McHits"]["origin"]
    if inactive_du:
        # only hits in active line
        origin = origin[blob["McHits"]["du"] != inactive_du]
    # get how many mchits were produced per muon in the bundle
    origin_dict = dict(zip(*np.unique(origin, return_counts=True)))
    return np.array([origin_dict.get(i, 0) for i in ids])


def get_mc_index(aanet_filename):
    # e.g. mcv5.40.mupage_10G.sirene.jterbr00005782.jorcarec.aanet.365.h5
    return int(aanet_filename.split(".")[-2])



import os
from pathlib import Path
import toml
import orcasong.core
import orcasong.extractors as extractors

# built-in extractors. First argument has to be the input filename,
# other parameters can be set via '[extractor_config]' in the toml
EXTRACTORS = {
    "nu_chain_neutrino": extractors.get_neutrino_mc_info_extr,
    "nu_chain_muon": extractors.get_muon_mc_info_extr,
    "nu_chain_noise": extractors.get_random_noise_mc_info_extr,
    "nu_chain_data": extractors.get_real_data_info_extr,
    "bundle_mc": extractors.BundleMCExtractor,
    "bundle_data": extractors.BundleDataExtractor,
}

MODES = {
    "graph": orcasong.core.FileGraph,
    "image": orcasong.core.FileBinner,
}


def add_parser_run(subparsers):
    parser = subparsers.add_parser(
        "run", description="Produce a dl file from an aanet file."
    )
    parser.add_argument("infile", type=str, help="Aanet file in h5 format.")
    parser.add_argument(
        "config",
        type=str,
        help=(
            "Orcasong configuration in toml format. Use prefix 'orcasong:' to load "
            "a toml from OrcaSong/configs."
        ),
    )
    parser.add_argument(
        "--detx",
        type=str,
        default=None,
        help=("Optional detx file to calibrate on the fly."),
    )
    parser.add_argument(
        "--outfile",
        type=str,
        default=None,
        help=("Path to output file. Default: Save with auto generated name in cwd."),
    )
    parser.set_defaults(func=run_orcasong)


def run_orcasong(infile, config, detx=None, outfile=None):
    setup_processor(infile, config, detx).run(infile=infile, outfile=outfile)


def setup_processor(infile, toml_file, detx_file=None):
    if toml_file.startswith("orcasong:"):
        toml_file = _get_config(toml_file[9:])
    cfg = toml.load(toml_file)
    processor = _get_verbose(cfg.pop("mode"), MODES)

    if "extractor" in cfg:
        extractor_cfg = cfg.pop("extractor_config", {})
        extractor = _get_verbose(cfg.pop("extractor"), EXTRACTORS)(
            infile, **extractor_cfg
        )
    else:
        extractor = None

    return processor(
        det_file=detx_file,
        extractor=extractor,
        **cfg,
    )


def _get_verbose(key, d):
    if key not in d:
        raise KeyError(f"Unknown key '{key}' (available: {list(d.keys())})")
    return d[key]


def _get_config(filename):
    direc = os.path.join(Path(orcasong.core.__file__).parents[1], "configs")
    files = {file: os.path.join(direc, file) for file in os.listdir(direc)}
    return _get_verbose(filename, files)



from unittest import TestCase
import numpy as np
import orcasong.extractors.bundles as bundles


class TestGetPlanePositions(TestCase):
    def setUp(self) -> None:
        self.positions = np.array(
            [
                [0, 0, 0],
                [0, 0, 5],
                [5, 3, 2],
                [0, 0, 2],
            ]
        )
        self.directions = np.array(
            [
                [0, 0, 1],
                [0, 0, 1],
                [0, 0, 1],
                [0, 1, -1],
            ]
        )

    def test_positions_flat_plane(self):
        plane_point = np.array([0, 0, 0])
        plane_normal = np.array([0, 0, 1])
        result = bundles.get_plane_positions(
            self.positions,
            self.directions,
            plane_point,
            plane_normal,
        )
        target = np.array(
            [
                [0, 0],
                [0, 0],
                [5, 3],
                [0, 2],
            ]
        )
        np.testing.assert_array_equal(result, target)

    def test_positions_flat_plane_no_plane_normal(self):
        plane_point = np.array([0, 0, 0])
        with self.assertRaises(ValueError):
            bundles.get_plane_positions(
                self.positions,
                self.directions,
                plane_point,
            )

    def test_positions_shifted_plane(self):
        plane_point = np.array([1, 0, 0])
        plane_normal = np.array([0, 0, 1])
        result = bundles.get_plane_positions(
            self.positions,
            self.directions,
            plane_point,
            plane_normal,
        )
        target = np.array(
            [
                [-1, 0],
                [-1, 0],
                [4, 3],
                [-1, 2],
            ]
        )
        np.testing.assert_array_equal(result, target)

    def test_positions_angled_plane(self):
        plane_point = np.array([0, 0, 0])
        plane_normal = np.array([1, 0, 1])
        result = bundles.get_plane_positions(
            self.positions,
            self.directions,
            plane_point,
            plane_normal,
        )
        target = np.array(
            [
                [0, 0],
                [0, 0],
                [np.sqrt(50), 3],
                [0, 2],
            ]
        )
        np.testing.assert_array_equal(result, target)

    def test_positions_angled_plane_2(self):
        result = bundles.get_plane_positions(
            positions=np.array([[-1, 0, 1]]),
            directions=np.array([[1, 0, -1]]),
            plane_point=np.array([0, 0, 0]),
            plane_normal=np.array([-1, 0, 1]),
        )
        np.testing.assert_array_equal(result, np.array([[0, 0]]))


class TestPairwiseDistances(TestCase):
    def setUp(self) -> None:
        self.positions_plane = np.array(
            [
                [0, -3],
                [4, 0],
                [0, 3],
            ]
        )

    def test_matrix(self):
        result = bundles.get_pairwise_distances(
            positions_plane=self.positions_plane,
            as_matrix=True,
        )
        target = np.array(
            [
                [0, 5, 6],
                [5, 0, 5],
                [6, 5, 0],
            ]
        )
        np.testing.assert_array_equal(result, target)

    def test_flat(self):
        result = bundles.get_pairwise_distances(
            positions_plane=self.positions_plane,
        )
        target = np.array([5, 6, 5])
        np.testing.assert_array_equal(result, target)
